import pkgutil
from importlib.util import find_spec

from .constants import DEFAULT_PAD_TOKEN, IGNORE_INDEX, SPACE_TOKENIZERS
from .data_model import (
    Serializable,
    sortedtuple,
    stringify_tuple_keys,
    summed_stat_dc,
    undefaultdict,
)
from .log_util import add_file_handler, log
from .multiproc import apply_args_and_kwargs, run_in_parallel, starmap_with_kwargs
from .sql import merge_databases
from .python import (
    all_equal,
    anyin,
    approx_list_split,
    approx_number_split,
    check_if_in_other_list,
    chunk_list,
    count_sublist_occurrences,
    dict_percentages,
    dict_without,
    flatten,
    first_true,
    identity,
    index_of_sublist_match,
    count_sublist_matches,
    list_find_new,
    list_flip,
    list_in_list,
    list_multiply,
    list_rindex,
    list_split,
    list_split_at_value,
    list_squeeze,
    list_get,
    make_list_unique,
    multify_text,
    naive_regex_escape,
    nth_index,
    re_line_matches,
    recursed_dict_percentages,
    recursed_merge_percent_stats,
    recursed_sum_up_stats,
    remove_until,
    removesuffixes,
    shortest_common_supersequence,
    split_multi,
    str_find_all,
    transpose_li_of_dict,
    transpose_list,
    unique_n_times,
    update_running_avg,
    dictsum,
    vector_sum,
)
from .json_util import FlexibleJSONEncoder
from .tools import bulk_rename, random_string
from .types_util import T, describe_type
from .os_tools import search_file_in_folder, find_all_file_paths
from .statistics import (
    Statlogger,
    Welfords,
    compute_metric,
    monte_carlo_bernoulli_p_value,
)

if find_spec("matplotlib") is not None:
    from .plotting import barplot_annotate_brackets

if find_spec("requests") is not None:
    from .llm_api import (
        process_file,
        process_folder,
        SglangModelWrapper,
        count_fulfilled_requests,
    )
if find_spec("scipy") is not None:
    from .statistics_scipy import (
        clopper_pearson_interval,
        wald_proportion_diff,
        double_diff_p_value,
        wald_proportion_p_value,
        exact_bernoulli_p_value,
    )
if find_spec("yaml") is not None:
    from .yaml_util import LiteralDumper, yaml_pretty_dump

if find_spec("torch") is not None and find_spec("numpy") is not None:
    from .print_tools import describe_array, describe_list

if find_spec("pandera") is not None:
    from .pandera_util import empty_dataframe_from_model

if find_spec("pandas") is not None:
    from .pandas_util import expand_json_cols

if find_spec("openai") is not None and find_spec("pydantic") is not None:
    from .openai_api import (
        AzureModelWrapper,
        ModelWrapper,
        human_readable_parse,
        local_image_to_data_url,
    )

if find_spec("pydantic") is not None and find_spec("yaml") is not None:
    from .configuration import from_file

if find_spec("torch") is not None:
    from .torch_helpers import (
        deserialize_tensor,
        disable_gradients,
        find_all_subarray_poses,
        free_cuda_memory,
        pad_along_dimension,
        print_memory_info,
        rolling_window,
        serialize_tensor,
        tensor_in,
    )

    if find_spec("accelerate") is not None:
        from .accelerate_tools import gather_dict

    if find_spec("transformers") is not None:
        from .model import (
            find_all_linear_names,
            get_max_memory_and_device_map,
            print_parameters_by_dtype,
            print_trainable_parameters,
            smart_tokenizer_and_embedding_resize,
        )
        from .training import train_eval_and_get_metrics
        from .transformer import (
            DataCollatorWithPadding,
            TokenStoppingCriteria,
            batch_tokenization,
            compute_seq_log_probability,
            dict_from_chat_template,
            find_tokens_with_str,
            flat_encode,
            generate_different_sequences,
            load_tk_with_pad_tk,
            obtain_offsets,
            regex_tokens_using_offsets,
            tokenize,
            tokenize_instances,
            transform_with_offsets,
            untokenize,
        )

        if find_spec("datasets") is not None:
            from .dataset import colorcode_dataset, colorcode_entry, describe_dataset

        if find_spec("transformer_heads") is not None:
            if find_spec("peft") is not None:
                from .peft_util import load_maybe_peft_model_tokenizer
            from .evaluation import (
                EvaluateFirstStepCallback,
                compute_classification_head_metrics,
                compute_metrics_functions,
            )

__version__ = "0.0.16"



import pandas as pd


def expand_json_cols(df: pd.DataFrame, cols: list[str] | str) -> pd.DataFrame:
    if isinstance(cols, str):
        cols = [cols]
    for col in cols:
        df = pd.concat([df, pd.json_normalize(df[col])], axis=1)
        df = df.drop(columns=[col])
    return df



from accelerate.utils import gather_object


def gather_dict(d, strict=False):
    out_dic = {key: [] for key in d}
    gathered_keys = gather_object(list(d.keys()))
    gathered_values = gather_object(list(d.values()))

    for key, value in zip(gathered_keys, gathered_values):
        if key in out_dic:
            out_dic[key].append(value)
        else:
            if strict:
                raise ValueError(f"Key {key} not found in original dictionary")

    return out_dic



import requests
import os
import json
import argparse
from urllib.parse import urljoin


class SglangModelWrapper:
    def __init__(self, host: str, model_name: str):
        self.host = host
        self.model_name = model_name

    def complete(self, messages: list[dict[str, str]], **kwargs) -> str:
        response = requests.post(
            urljoin(self.host, "/v1/chat/completions"),
            json={"model": self.model_name, "messages": messages, **kwargs},
        )
        return response.json()["choices"][0]["message"]["content"]


def process_file(
    file_path: str,
    model_name: str,
    url: str = "http://localhost:30000/v1/chat/completions",
    temperature: float = 1.0,
    max_tokens: int = 100,
):

    with open(file_path, "r") as f:
        data = [json.loads(x) for x in f.readlines()]

    made_changes = False
    for example in data:
        if "response" not in example:
            made_changes = True
            llm_request = {
                "model": model_name,
                "messages": [{"role": "user", "content": example["prompt"]}],
                "max_tokens": max_tokens,
                "temperature": temperature,
            }
            header = {
                "Content-Type": "application/json",
            }
            result = requests.post(url, json=llm_request, headers=header)
            example["response"] = result.json()["choices"][0]["message"]["content"]

    if made_changes:
        with open(file_path, "w") as f:
            for example in data:
                f.write(json.dumps(example) + "\n")


def process_folder(
    folder_path: str,
    model_name: str,
    url: str = "http://localhost:30000/v1/chat/completions",
    temperature: float = 1.0,
    max_tokens: int = 100,
):
    for file in os.listdir(folder_path):
        if file.endswith(".jsonl"):
            process_file(
                os.path.join(folder_path, file),
                model_name,
                url,
                temperature,
                max_tokens,
            )


def count_fulfilled_requests(
    folder_path: str,
):
    fulfilled = 0
    todo = 0
    for file in os.listdir(folder_path):
        if file.endswith(".jsonl"):
            with open(os.path.join(folder_path, file), "r") as f:
                data = [json.loads(x) for x in f.readlines()]
            for example in data:
                if "response" in example:
                    fulfilled += 1
                else:
                    todo += 1
    print(f"Fulfilled: {fulfilled}, Todo: {todo}")


def do_process_folder():
    parser = argparse.ArgumentParser()
    parser.add_argument("folder_path", type=str)
    parser.add_argument("model_name", type=str)
    parser.add_argument(
        "--url", type=str, default="http://localhost:30000/v1/chat/completions"
    )
    parser.add_argument("--temperature", type=float, default=1.0)
    parser.add_argument("--max_tokens", type=int, default=100)
    args = parser.parse_args()
    process_folder(
        args.folder_path, args.model_name, args.url, args.temperature, args.max_tokens
    )


def do_process_file():
    parser = argparse.ArgumentParser()
    parser.add_argument("file_path", type=str)
    parser.add_argument("model_name", type=str)
    parser.add_argument(
        "--url", type=str, default="http://localhost:30000/v1/chat/completions"
    )
    parser.add_argument("--temperature", type=float, default=1.0)
    parser.add_argument("--max_tokens", type=int, default=100)
    args = parser.parse_args()
    process_file(
        args.file_path, args.model_name, args.url, args.temperature, args.max_tokens
    )



import inspect
import logging
import os
from logging.handlers import RotatingFileHandler

if not "log_util" in logging.Logger.manager.loggerDict:
    logger = logging.getLogger("log_util")
    logger.setLevel(logging.DEBUG)
    hl = logging.StreamHandler()
    hl.setLevel(logging.INFO)
    hl.setFormatter(logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S"))
    logger.addHandler(hl)

level_map = {
    "WARN": logging.WARNING,
    "INFO": logging.INFO,
    "ERROR": logging.ERROR,
    "DEBUG": logging.DEBUG,
    "WARNING": logging.WARNING,
}


def add_file_handler(file_path, level=logging.INFO, max_bytes=None):
    if isinstance(level, str):
        level = level_map[level.upper()]
    if max_bytes:
        fl = RotatingFileHandler(file_path, maxBytes=max_bytes, backupCount=0, mode="w")
    else:
        fl = logging.FileHandler(file_path, mode="w")
    fl.setLevel(level)
    fl.setFormatter(logging.Formatter("%(asctime)s %(message)s", datefmt="%H:%M:%S"))
    logger.addHandler(fl)


def log(*messages, level=logging.INFO):
    from termcolor import colored

    if type(level) == str:
        level = level_map[level.upper()]
    caller = inspect.stack()[1]
    msg = f'{os.path.basename(caller.filename)}:{caller.lineno} {" ".join(map(str, messages))}'
    if level == logging.ERROR:
        msg = colored(msg, "red")
    elif level == logging.WARNING:
        msg = colored(msg, "yellow")
    logger.log(level, msg)



from collections import defaultdict
from copy import copy
from dataclasses import asdict, is_dataclass
from typing import Callable

from ykutil.python import recursed_merge_percent_stats, recursed_sum_up_stats
from ykutil.types_util import T


def summed_stat_dc(
    datas: list[T], avg_keys: tuple = (), weight_attr="num_examples"
) -> T:
    weights = [getattr(d, weight_attr, 1) for d in datas]
    return datas[0].__class__(
        **{
            key: (
                recursed_merge_percent_stats(
                    [getattr(d, key) for d in datas], weights=weights
                )
                if key in avg_keys
                else recursed_sum_up_stats([getattr(d, key) for d in datas])
            )
            for key in datas[0].__dict__
        }
    )


def undefaultdict(d: dict | object, do_copy=False):
    # Fix related to https://bugs.python.org/issue35540
    if isinstance(d, defaultdict):
        d = dict(d)
    elif do_copy:
        d = copy(d)
    if is_dataclass(d):
        for key, value in d.__dict__.items():
            if isinstance(value, dict):
                setattr(d, key, undefaultdict(value, do_copy=do_copy))
    else:
        if isinstance(d, dict):
            for key, value in d.items():
                d[key] = undefaultdict(value, do_copy=do_copy)
    return d


def stringify_tuple_keys(d: dict | object):
    k = d.__dict__ if is_dataclass(d) else d
    for key, value in list(k.items()):
        if isinstance(value, dict) or is_dataclass(value):
            k[key] = stringify_tuple_keys(value)
        if isinstance(key, tuple):
            del k[key]
            k[str(key)] = value
    return d


class Serializable(dict):
    """Inherit from this when creating a dataclass to make it
    json serializable and hashable.
    """

    def __hash__(self):
        return hash(tuple(sorted(self.items())))

    def items(self):
        return asdict(self).items()

    def __len__(self):
        return len(asdict(self))

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result


def sortedtuple(sort_fun: Callable, fixed_len=None):
    def __new__(cls, *args):
        assert fixed_len is None or len(args) == fixed_len
        return tuple.__new__(cls, sorted(args, key=sort_fun))

    return type("SortedTuple", (tuple,), {"__new__": __new__})



import sqlite3
from ykutil import log


def merge_databases(target_db: str, source_dbs: list[str]):
    """
    Merge multiple SQLite databases into one.

    Parameters:
    target_db (str): Path to the target SQLite database.
    source_dbs (list): List of paths to source SQLite databases.
    """
    # Create a new database or connect to an existing one
    target_conn = sqlite3.connect(target_db)
    target_cursor = target_conn.cursor()

    for source_db in source_dbs:
        print(f"Merging {source_db} into {target_db}...")

        # Connect to the source database
        source_conn = sqlite3.connect(source_db)
        source_cursor = source_conn.cursor()

        # Get all table names from the source database
        source_cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = [row[0] for row in source_cursor.fetchall()]

        for table in tables:
            print(f"  Copying table {table}...")

            # Fetch column names
            source_cursor.execute(f"PRAGMA table_info({table})")
            columns = [row[1] for row in source_cursor.fetchall()]
            column_list = ", ".join(columns)

            # Copy data from source to target
            source_cursor.execute(f"SELECT {column_list} FROM {table}")
            rows = source_cursor.fetchall()

            # Ensure the target table exists
            target_cursor.execute(
                f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table}';"
            )
            if not target_cursor.fetchone():
                print(f"  Creating table {table} in target database...")
                source_cursor.execute(
                    f"SELECT sql FROM sqlite_master WHERE type='table' AND name='{table}'"
                )
                create_table_sql = source_cursor.fetchone()[0]
                target_cursor.execute(create_table_sql)

            # Insert data into the target database
            placeholders = ", ".join(["?"] * len(columns))
            target_cursor.executemany(
                f"INSERT INTO {table} ({column_list}) VALUES ({placeholders})", rows
            )

        # Commit and close the source connection
        target_conn.commit()
        source_conn.close()

    # Close the target connection
    target_conn.close()
    print("Merge completed successfully.")



from scipy.stats import beta, norm
from math import sqrt
from scipy.integrate import dblquad


def clopper_pearson_interval(successes, trials, confidence_level=0.95):
    """
    Compute the Clopper-Pearson confidence interval.

    Parameters:
        successes (int): Number of successes.
        trials (int): Total number of trials.
        confidence_level (float): Confidence level (e.g., 0.95 for 95% confidence).

    Returns:
        (lower_bound, upper_bound): The confidence interval bounds.
    """
    assert successes <= trials
    alpha = 1 - confidence_level
    lower_bound = (
        beta.ppf(alpha / 2, successes, trials - successes + 1) if successes > 0 else 0.0
    )
    upper_bound = (
        beta.ppf(1 - alpha / 2, successes + 1, trials - successes)
        if successes < trials
        else 1.0
    )
    return lower_bound, upper_bound


def wald_proportion_diff(s1: int, t1: int, s2: int, t2: int, ci: float = 0.95):
    """Given samples of two bernoulli distributions, compute the Wald test statistic.

    >>> wald_proportion_diff(70, 100, 50, 100)
    (0.06706877501808317, 0.33293122498191674)
    """

    p1 = s1 / t1
    p2 = s2 / t2

    se = sqrt(p1 * (1 - p1) / t1 + p2 * (1 - p2) / t2)

    z = norm.ppf(1 - (1 - ci) / 2)

    interval = (p1 - p2 - z * se, p1 - p2 + z * se)

    return interval


def exact_bernoulli_p_value(s1: int, t1: int, s2: int, t2: int):
    """Given samples of two bernoulli distributions, compute the exact p-value.
    Bayesian approach using beta function integration

    >>> exact_bernoulli_p_value(70, 100, 50, 100)
    0.001979681257445076
    """
    # Beta values for Bayesian estimation
    f1 = t1 - s1 + 1
    f2 = t2 - s2 + 1
    s1 = s1 + 1
    s2 = s2 + 1

    # Define the joint probability density function for Beta distributions
    def joint_beta_pdf(p1, p2, s1, f1, s2, f2):
        return beta.pdf(p1, s1, f1) * beta.pdf(p2, s2, f2)

    # Compute P(p1 > p2) via numerical integration
    prob_integral, _ = dblquad(
        lambda p2, p1: joint_beta_pdf(p1, p2, s1, f1, s2, f2),
        0,
        1,  # p1 bounds
        lambda p1: 0,
        lambda p1: p1,  # p2 bounds (p2 < p1)
    )

    return 1 - prob_integral


def wald_proportion_p_value(s1: int, t1: int, s2: int, t2: int):
    """Given samples of two bernoulli distributions, compute the Wald test p-value.
    Frequentist approach

    >>> wald_proportion_p_value(70, 100, 50, 100)
    0.003189699706216853
    """
    p1 = s1 / t1
    p2 = s2 / t2

    se = sqrt(p1 * (1 - p1) / t1 + p2 * (1 - p2) / t2)

    z = (p1 - p2) / se

    p_value = 2 * (1 - norm.cdf(abs(z)))

    return p_value


def double_diff_p_value(
    s1: int, t1: int, s2: int, t2: int, s3: int, t3: int, s4: int, t4: int
):
    """Given successes s1,s2,s3,s4 of bernoulli distributions of totals t1,t2,t3,t4
    and probabilities p1,p2,p3,p4, compute the p-value that (p1-p2)>(p3-p4)
    """
    p1_hat = s1 / t1
    p2_hat = s2 / t2
    p3_hat = s3 / t3
    p4_hat = s4 / t4

    D1_hat = p1_hat - p2_hat
    D2_hat = p3_hat - p4_hat

    var_D1 = (p1_hat * (1 - p1_hat) / t1) + (p2_hat * (1 - p2_hat) / t2)
    var_D2 = (p3_hat * (1 - p3_hat) / t3) + (p4_hat * (1 - p4_hat) / t4)

    var_X = var_D1 + var_D2
    std_X = sqrt(var_X)

    Z = (D1_hat - D2_hat) / std_X

    p_value = 1 - norm.cdf(Z)

    return p_value


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import time
import threading
from collections import defaultdict
from math import sqrt
import numpy as np


class Statlogger:
    def __init__(self):
        self.statistics = defaultdict(float)
        self.sum_stats = defaultdict(float)
        self.statnumbers = defaultdict(int)
        self.timers = dict()

    def reset(self):
        self.statistics.clear()
        self.sum_stats.clear()
        self.statnumbers.clear()

    def update(self, key, value):
        self.statistics[key] = (
            (self.statistics[key] * self.statnumbers[key]) + value
        ) / (self.statnumbers[key] + 1)
        self.statnumbers[key] = self.statnumbers[key] + 1

    def update_sum_stat(self, key, value):
        self.sum_stats[key] += value

    def start_timer(self, key):
        thread_id = threading.get_ident()
        thread_key = f"{key}_{thread_id}"
        self.timers[thread_key] = time.perf_counter()

    def stop_timer(self, key, average=False, summed=True):
        thread_id = threading.get_ident()
        thread_key = f"{key}_{thread_id}"
        assert summed or average
        if average:
            self.update(f"average_{key}", time.perf_counter() - self.timers[thread_key])
        if summed:
            self.update_sum_stat(
                f"total_{key}", time.perf_counter() - self.timers[thread_key]
            )
        del self.timers[thread_key]

    @property
    def stats(self):
        some_dic = dict(self.statistics)
        some_dic.update(self.sum_stats)
        return some_dic


class Welfords:
    def __init__(self):
        self.count: int = 0
        self.mean: float = 0.0
        self.M2: float = 0.0

    def update(self, new_value):
        self.count += 1
        delta = new_value - self.mean
        self.mean += delta / self.count
        delta2 = new_value - self.mean
        self.M2 += delta * delta2

    @property
    def std(self) -> float:
        return sqrt(self.M2 / self.count) if self.count > 1 else 0.0

    def normalize(self, value):
        return (value - self.mean) / self.std if self.std > 0 else 0.0

    def unnormalize(self, value):
        return value * self.std + self.mean

    def __repr__(self):
        return f"Welfords(Mean: {self.mean}, Std: {self.std}, Count: {self.count})"


def compute_metric(pred: float, targ: float, metric_type: str):
    pred = float(pred)
    targ = float(targ)
    match metric_type:
        case "mae":
            return abs(pred - targ)
        case "mse":
            return (pred - targ) ** 2
        case "acc":
            return int(round(pred) == targ)
        case "bce":
            return -targ * np.log(pred) - (1 - targ) * np.log(1 - pred)
        case _:
            raise ValueError(f"Unknown metric type {metric_type}")


def monte_carlo_bernoulli_p_value(s1: int, t1: int, s2: int, t2: int, n: int = 10000):
    """Given samples of two bernoulli distributions, compute the monte-carlo p-value.
    Bayesian approach using sampling

    >>> np.random.seed(42)
    >>> 0.001 < monte_carlo_bernoulli_p_value(70, 100, 50, 100) < 0.003
    True
    """
    # Beta values for Bayesian estimation
    f1 = t1 - s1 + 1
    f2 = t2 - s2 + 1
    s1 = s1 + 1
    s2 = s2 + 1

    # Draw samples from the Beta distributions
    samples_p1 = np.random.beta(s1, f1, n)
    samples_p2 = np.random.beta(s2, f2, n)

    # Compute probability that p1 > p2
    prob_p1_greater_p2 = np.mean(samples_p1 > samples_p2)

    return 1 - prob_p1_greater_p2


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import json
import os
from typing import Optional

import torch
from peft import AutoPeftModelForCausalLM
from transformer_heads import load_lora_with_heads
from transformer_heads.util.helpers import get_model_params
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
)


def load_unsloth_model_for_inference(
    model_path, torch_dtype=torch.bfloat16, load_in_4bit=False, device_map="auto"
):
    from unsloth import FastLanguageModel

    model, tokenizer = FastLanguageModel.from_pretrained(
        model_name=model_path,
        max_seq_length=4096,
        dtype=torch_dtype,
        load_in_4bit=load_in_4bit,
        device_map=device_map,
    )
    FastLanguageModel.for_inference(model)

    tokenizer.pad_token_id = tokenizer.eos_token_id

    return model, tokenizer


def load_maybe_peft_model_tokenizer(
    model_path,
    device_map="auto",
    quantization_config: Optional[BitsAndBytesConfig] = None,
    attention_implementation="eager",
    only_inference=True,
    torch_dtype=torch.bfloat16,
    use_unsloth=False,
) -> tuple[PreTrainedModel, PreTrainedTokenizer]:
    extra_args = {}
    if quantization_config is not None:
        extra_args["quantization_config"] = quantization_config
    if os.path.exists(os.path.join(model_path, "adapter_config.json")):
        with open(os.path.join(model_path, "adapter_config.json")) as f:
            base_model_name = json.load(f)["base_model_name_or_path"]
        if use_unsloth:
            print("Loading fast unsloth model", model_path)
            return load_unsloth_model_for_inference(
                model_path,
                torch_dtype=torch_dtype,
                device_map=device_map,
                load_in_4bit=quantization_config is not None
                and quantization_config.load_in_4bit,
            )
        elif os.path.isfile(os.path.join(model_path, "head_configs.json")):
            model_params = get_model_params(base_model_name)
            model = load_lora_with_heads(
                base_model_class=model_params["model_class"],
                path=model_path,
                quantization_config=quantization_config,
                only_inference=only_inference,
                attn_implementation=attention_implementation,
                torch_dtype=torch_dtype,
                device_map=device_map,
            )
        else:
            model = AutoPeftModelForCausalLM.from_pretrained(
                model_path,
                device_map=device_map,
                attn_implementation=attention_implementation,
                torch_dtype=torch_dtype,
                **extra_args,
            )
    else:
        model = AutoModelForCausalLM.from_pretrained(
            model_path,
            device_map=device_map,
            attn_implementation=attention_implementation,
            torch_dtype=torch_dtype,
            **extra_args,
        )
        base_model_name = model_path

    # For some reason, the instruct model got the wrong eos_token here
    # https://huggingface.co/meta-llama/Meta-Llama-3-8B-Instruct/commit/a8977699a3d0820e80129fb3c93c20fbd9972c41
    # So we load the base model instead
    base_model_name = base_model_name.replace("-Instruct", "")

    tokenizer = AutoTokenizer.from_pretrained(base_model_name)
    if tokenizer.pad_token_id is None:
        tokenizer.pad_token = tokenizer.eos_token
    return model, tokenizer



import os
import re
import random
import string


def bulk_rename(directory, pattern, replacement):
    """
    Renames files in the specified directory by replacing occurrences of the given
    pattern in the filename with the replacement string.

    Parameters:
        directory (str): The path to the folder containing the files to rename.
        pattern (str): The regex pattern to match in filenames.
        replacement (str): The string to replace the matched pattern.
    """

    # Check if the directory exists
    if not os.path.isdir(directory):
        print(f"Error: Directory '{directory}' does not exist.")
        return

    # Compile the regex pattern
    regex = re.compile(pattern)

    # Iterate over all files in the directory
    for filename in os.listdir(directory):
        # Only match files that fit the specified pattern
        if regex.search(filename):
            # Construct the new filename
            new_filename = regex.sub(replacement, filename)

            # Get the full paths for renaming
            old_path = os.path.join(directory, filename)
            new_path = os.path.join(directory, new_filename)

            # Rename the file
            try:
                os.rename(old_path, new_path)
                print(f"Renamed '{filename}' to '{new_filename}'")
            except Exception as e:
                print(f"Error renaming '{filename}': {e}")


def random_string(length, capital=True, numbers=False, special=False):
    """
    Generates a random string of the specified length.

    Parameters:
        length (int): The length of the random string to generate.
        capital (bool): If True, include uppercase letters. Default is True.
        numbers (bool): If True, include digits. Default is False.
        special (bool): If True, include special characters. Default is False.

    Returns:
        str: The generated random string.
    """
    characters = string.ascii_lowercase
    if capital:
        characters += string.ascii_uppercase
    if numbers:
        characters += string.digits
    if special:
        characters += string.punctuation

    return "".join(random.choice(characters) for _ in range(length))



from itertools import repeat
from math import ceil
from multiprocessing import Pool
from typing import Callable


def starmap_with_kwargs(pool, fn, args_iter, kwargs_iter):
    args_for_starmap = zip(repeat(fn), args_iter, kwargs_iter)
    return pool.starmap(apply_args_and_kwargs, args_for_starmap)


def apply_args_and_kwargs(fn, args, kwargs):
    return fn(*args, **kwargs)


def run_in_parallel(
    func: Callable,
    list_ordered_kwargs: dict[list],
    num_workers: int,
    extra_kwargs: dict = {},
):
    """Processes the list_ordered args with parallel processes."""
    num_elems = len(next(iter(list_ordered_kwargs.values())))
    chunk_len = ceil(num_elems / num_workers)
    args_iter = repeat([])
    kwargs_iter = []
    for start, end in zip(
        range(0, num_elems, chunk_len),
        range(chunk_len, num_elems + chunk_len, chunk_len),
    ):
        res = {key: value[start:end] for key, value in list_ordered_kwargs.items()}
        res.update({key: value for key, value in extra_kwargs.items()})
        kwargs_iter.append(res)

    with Pool(num_workers) as pool:
        res = starmap_with_kwargs(pool, func, args_iter, kwargs_iter)
    return res



import matplotlib.pyplot as plt


def barplot_annotate_brackets(
    num1,
    num2,
    data,
    center,
    height,
    yerr=None,
    dh=0.05,
    barh=0.05,
    fs=None,
    maxasterix=None,
):
    """
    Annotate barplot with p-values.

    :param num1: number of left bar to put bracket over
    :param num2: number of right bar to put bracket over
    :param data: string to write or number for generating asterixes
    :param center: centers of all bars (like plt.bar() input)
    :param height: heights of all bars (like plt.bar() input)
    :param yerr: yerrs of all bars (like plt.bar() input)
    :param dh: height offset over bar / bar + yerr in axes coordinates (0 to 1)
    :param barh: bar height in axes coordinates (0 to 1)
    :param fs: font size
    :param maxasterix: maximum number of asterixes to write (for very small p-values)
    """

    if type(data) is str:
        text = data
    else:
        # * is p < 0.05
        # ** is p < 0.005
        # *** is p < 0.0005
        # etc.
        text = ""
        p = 0.05

        while data < p:
            text += "*"
            p /= 10.0

            if maxasterix and len(text) == maxasterix:
                break

        if len(text) == 0:
            text = "n. s."

    lx, ly = center[num1], height[num1]
    rx, ry = center[num2], height[num2]

    if yerr:
        ly += yerr[num1]
        ry += yerr[num2]

    ax_y0, ax_y1 = plt.gca().get_ylim()
    dh *= ax_y1 - ax_y0
    barh *= ax_y1 - ax_y0

    print(ly, ry, dh)
    print(center, height, yerr)
    y = max(ly, ry) + dh

    barx = [lx, lx, rx, rx]
    bary = [y, y + barh, y + barh, y]
    mid = ((lx + rx) / 2, y + barh)

    plt.plot(barx, bary, c="black")

    kwargs = dict(ha="center", va="bottom")
    if fs is not None:
        kwargs["fontsize"] = fs

    plt.text(*mid, text, **kwargs)



from collections import defaultdict
import os
from typing import Type

import yaml
from pydantic import BaseModel

from ykutil.types_util import T


def from_file(cls: Type[T], config_file: str, **argmod) -> tuple[T, dict]:
    with open(config_file, "r") as f:
        config = defaultdict(dict, yaml.safe_load(f.read()))

    for key, value in tuple(argmod.items()):
        path = key.split(".")
        c = config
        for p in path[:-1]:
            if p not in c:
                if isinstance(c, list):
                    p = int(p)
                else:
                    break
            c = c[p]
        else:
            if path[-1] in c:
                c[path[-1]] = value
                del argmod[key]

    if issubclass(cls, BaseModel):
        args = cls.model_validate(config)

    else:
        from dacite import from_dict

        args = from_dict(data_class=cls, data=config)
    return args, argmod
